@@ -21,6 +21,7 @@ proc call_page {{useenv ""} {createinterp 0}} {
 		set inchan [lindex $env(RIVET_INTERFACE) 1]
 		set outchan [lindex $env(RIVET_INTERFACE) 2]
 		set elogchan [lindex $env(RIVET_INTERFACE) 3]
+		array set headers [lindex $env(RIVET_INTERFACE) 4]
 	}
 
 	# Determine if a sub-file has been requested
@@ -134,6 +135,7 @@ proc call_page {{useenv ""} {createinterp 0}} {
 			if {$createinterp} {
 				set myinterp [interp create]
 
+				$myinterp eval [list set ::auto_path $::auto_path]
 				$myinterp eval [list package require tclrivet]
 				$myinterp eval [list unset -nocomplain ::env]
 				$myinterp eval [list array set ::env [array get env]]
@@ -161,14 +163,14 @@ proc call_page {{useenv ""} {createinterp 0}} {
 				} err]} {
 					$myinterp eval [list rivet_error]
 
-					$myinterp eval [list rivet_flush]
+					$myinterp eval [list rivet_flush 1]
 
 					interp delete $myinterp
 					return
 				}
 
 				# Flush the output stream
-				$myinterp eval [list rivet_flush]
+				$myinterp eval [list rivet_flush 1]
 
 				interp delete $myinterp
 			} else {
@@ -184,13 +186,13 @@ proc call_page {{useenv ""} {createinterp 0}} {
 				} err]} {
 					rivet_error
 
-					rivet_flush
+					rivet_flush 1
 
 					return
 				}
 
 				# Flush the output stream
-				rivet_flush
+				rivet_flush 1
 			}
 	
 			return
@@ -699,15 +701,17 @@ proc call_page {{useenv ""} {createinterp 0}} {
 	
 	# Dump static files
 	if {[info exists statictype]} {
-		::rivet::cgi_server_writehttpheader 200 [array get env]
+		set filelen 0
+		catch {
+			set filelen [file size $targetfile]
+		}
+
+		::rivet::cgi_server_writehttpheader 200 [array get env] $filelen
 		tcl_puts $outchan "Content-type: $statictype"
 		catch {
 			tcl_puts $outchan "Last-Modified: [clock format [file mtime $targetfile] -format {%a, %d %b %Y %H:%M:%S GMT} -gmt 1]"
 			tcl_puts $outchan "Expires: Tue, 19 Jan 2038 03:14:07 GMT"
 		}
-		catch {
-			tcl_puts $outchan "Content-Length: [file size $targetfile]"
-		}
 		tcl_puts $outchan ""
 	
 		set fd [open $targetfile r]
@@ -746,12 +750,59 @@ proc print_help {} {
 	tcl_puts "   --sslkey file      Path to our key (server mode)"
 	tcl_puts "   --sslcafile file   Path to CA certificate (server mode)"
 	tcl_puts "   --sslcadir dir     Path to directory containing CA certificates (server mode)"
-	tcl_puts "   --sslreqcert bool  Require client certificate (server mode, default is \"no\")"
+	tcl_puts "   --sslreqcert bool  Request client certificate (server mode, default is \"no\")"
 	tcl_puts "   --cgi              Execute as a CGI"
 	tcl_puts "   --help             This help"
 	tcl_puts "   --version          Print version and exit"
 }
 
+proc rivet_cgi_tls_callback {logfd elogfd mode args} {
+	switch -- $mode {
+		"error" {
+			foreach {chan msg} $args break
+
+			if {$elogfd != ""} {
+				tcl_puts $elogfd "($chan) TLS error: $msg"
+			}
+
+			return 1
+		}
+		"verify" {
+			foreach {chan depth cert rc err} $args break
+
+			# Only care about client certificates, all others follow normal rules
+			if {$depth != 0} {
+				return $rc
+			}
+
+			if {$rc != "1"} {
+				# Do not note that this certificate is valid, but return OK
+
+				if {$elogfd != ""} {
+					tcl_puts $elogfd "($chan) Invalid cert $cert: $err"
+				}
+
+				set ::rivet_cgi_tls_verified($chan) 0
+
+				return 1
+			}
+
+			if {![info exists ::rivet_cgi_tls_verified($chan)]} {
+				# Verified
+
+				set ::rivet_cgi_tls_verified($chan) 1
+			}
+
+			return 1
+		}
+		"info" {
+		}
+		default {
+			return -code error "bad option \"$mode\": must be one of error, info, or verify"
+		}
+	}
+}
+
 proc rivet_cgi_server {addr ports foreground initscp logfile errorlogfile maxthreads sslopts_array} {
 	catch {
 		package require Tclx
@@ -834,7 +885,7 @@ proc rivet_cgi_server {addr ports foreground initscp logfile errorlogfile maxthr
 
 			package require tls
 
-			set cmd [list tls::socket -server $servercmd -tls1 true]
+			set cmd [list tls::socket -server $servercmd -tls1 true -command [list rivet_cgi_tls_callback $logfd $elogfd]]
 			foreach opt [array names sslopts] {
 				set val $sslopts($opt)
 
@@ -921,6 +972,8 @@ proc rivet_cgi_server_request {hostport logfd elogfd pmodel maxthreads httpmode
 			set mypid [fork]
 			if {$mypid != 0} {
 				catch {
+					unset -nocomplain ::rivet_cgi_tls_verified($sock)
+
 					close $sock
 				}
 				return
@@ -954,7 +1007,10 @@ proc rivet_cgi_server_request {hostport logfd elogfd pmodel maxthreads httpmode
 				if {$numthreads > $maxthreads} {
 					tcl_puts $elogfd "Exceeded maximum number of threads ($maxthreads): $numthreads, closing socket!"
 
+					unset -nocomplain ::rivet_cgi_tls_verified($sock)
+
 					close $sock
+
 					return
 				}
 			}
@@ -975,12 +1031,16 @@ proc rivet_cgi_server_request {hostport logfd elogfd pmodel maxthreads httpmode
 				}
 
 				# Copy the appropriate variables to the new thread
-				foreach var [list ::starkit::topdir] {
+				foreach var [list ::starkit::topdir ::rivet_cgi_tls_verified($sock)] {
 					if {[namespace qualifiers $var] != ""} {
 						thread::send $threadId [list namespace eval [namespace qualifiers $var] ""]
 					}
 
-					thread::send $threadId [list set $var [set $var]]
+					if {[info exists $var]} {
+						thread::send $threadId [list set $var [set $var]]
+					} else {
+						thread::send $threadId [list unset -nocomplain $var]
+					}
 				}
 
 				# Copy the appropriate procedures to the new thread
@@ -1041,11 +1101,14 @@ proc rivet_cgi_server_request {hostport logfd elogfd pmodel maxthreads httpmode
 
 		# Handle connection from data
 		fconfigure $sock -buffering line
+		fconfigure $sock -translation crlf
 		fileevent $sock readable [list rivet_cgi_server_request_data $sock $addr $hostport $logfd $elogfd $pmodel]
 
 		vwait ::rivetstarkit::finished($sock)
 
 		catch {
+			unset -nocomplain ::rivet_cgi_tls_verified($sock)
+
 			close $sock
 		}
 	} err]} {
@@ -1090,11 +1153,19 @@ proc rivet_cgi_server_request_data {sock addr hostport logfd elogfd pmodel} {
 		"NEW" {
 			set sockinfo(requestline) $line
 			set work [split $line " "]
+
 			set sockinfo(method) [string toupper [lindex $work 0]]
-			set sockinfo(httpproto) [string toupper [lindex $work end]]
+
+			if {[llength $work] == 2} {
+				set sockinfo(url) [lindex $work 1]
+				set sockinfo(httpproto) "HTTP/1.0"
+			} else {
+				set sockinfo(url) [join [lrange $work 1 end-1] " "]
+				set sockinfo(httpproto) [string toupper [lindex $work end]]
+			}
+
 			set sockinfo(state) HEADERS
 
-			set sockinfo(url) [join [lrange $work 1 end-1] " "]
 			set sockinfo(url) [regsub {http://[^/]*/} $sockinfo(url) {/}]
 
 			set work [split $sockinfo(url) ?]
@@ -1106,6 +1177,9 @@ proc rivet_cgi_server_request_data {sock addr hostport logfd elogfd pmodel} {
 			# We only support GET and POST, everyone else we just close on.
 			if {$sockinfo(method) != "GET" && $sockinfo(method) != "POST"} {
 				set ::rivetstarkit::finished($sock) 1
+
+				unset -nocomplain ::rivet_cgi_tls_verified($sock)
+
 				close $sock
 			}
 		}
@@ -1122,7 +1196,11 @@ proc rivet_cgi_server_request_data {sock addr hostport logfd elogfd pmodel} {
 	}
 
 	if {$sockinfo(state) == "HANDLEREQUEST"} {
-		array set headers $sockinfo(headers)
+		if {[info exists sockinfo(headers)]} {
+			array set headers $sockinfo(headers)
+		} else {
+			array set headers [list]
+		}
 
 		# Write log entry
 		if {[info exists headers(USER-AGENT)]} {
@@ -1135,8 +1213,18 @@ proc rivet_cgi_server_request_data {sock addr hostport logfd elogfd pmodel} {
 
 		set localinfo [fconfigure $sock -sockname]
 		if {![info exists headers(CONNECTION)]} {
-			set headers(CONNECTION) "close"
+			switch -- $sockinfo(httpproto) {
+				"HTTP/1.1" {
+					set headers(CONNECTION) "keep-alive"
+				}
+				default {
+					set headers(CONNECTION) "close"
+				}
+			}
+		} else {
+			set headers(CONNECTION) [string tolower $headers(CONNECTION)]
 		}
+
 		if {![info exists headers(HOST)]} {
 			set headers(HOST) [lindex $localinfo 1]
 		}
@@ -1190,7 +1278,7 @@ proc rivet_cgi_server_request_data {sock addr hostport logfd elogfd pmodel} {
 
 		# Add Rivet Interface specification to fake environment, so further
 		# Rivet/CGI knows how to interface
-		set myenv(RIVET_INTERFACE) [list FULLHEADERS $sock $sock $elogfd]
+		set myenv(RIVET_INTERFACE) [list FULLHEADERS $sock $sock $elogfd [array get headers]]
 
 		# Set TLS Socket Info
 		array set tlsinfo_peer [list sbits 0]
@@ -1198,7 +1286,6 @@ proc rivet_cgi_server_request_data {sock addr hostport logfd elogfd pmodel} {
 		catch {
 			array set tlsinfo_peer [tls::status $sock]
 
-			set myenv(HTTPS) on
 		}
 		catch {
 			array set tlsinfo_local [tls::status -local $sock]
@@ -1206,7 +1293,18 @@ proc rivet_cgi_server_request_data {sock addr hostport logfd elogfd pmodel} {
 
 		## Set TLS client CGI Variables
 		if {$tlsinfo_peer(sbits) != 0} {
-			set myenv(SSL_CLIENT_VERIFY) SUCCESS
+			set myenv(HTTPS) on
+
+			if {![info exists ::rivet_cgi_tls_verified($sock)]} {
+				set ::rivet_cgi_tls_verified($sock) 0
+			}
+
+			if {$::rivet_cgi_tls_verified($sock) == "1"} {
+				set myenv(SSL_CLIENT_VERIFY) SUCCESS
+			} else {
+				unset -nocomplain myenv(SSL_CLIENT_VERIFY)
+			}
+
 			foreach {myenvvar tlsvar} [list SSL_CLIENT_S_DN subject SSL_CLIENT_I_DN issuer SSL_CLIENT_V_START notBefore SSL_CLIENT_V_END notAfter SSL_CLIENT_M_SERIAL serial SSL_CIPHER cipher SSL_CIPHER_USEKEYSIZE sbits] {
 				if {![info exists tlsinfo_peer($tlsvar)]} {
 					continue
@@ -1214,11 +1312,12 @@ proc rivet_cgi_server_request_data {sock addr hostport logfd elogfd pmodel} {
 
 				set myenv($myenvvar) $tlsinfo_peer($tlsvar)
 			}
-
 		}
 
 		## Set TLS server CGI Variables
 		if {$tlsinfo_local(sbits) != 0} {
+			set myenv(HTTPS) on
+
 			foreach {myenvvar tlsvar} [list SSL_SERVER_S_DN subject SSL_SERVER_I_DN issuer SSL_SERVER_V_START notBefore SSL_SERVER_V_END notAfter SSL_SERVER_M_SERIAL serial SSL_CIPHER cipher SSL_CIPHER_USEKEYSIZE sbits] {
 				if {![info exists tlsinfo_local($tlsvar)]} {
 					continue
@@ -1265,10 +1364,10 @@ proc rivet_cgi_server_request_data {sock addr hostport logfd elogfd pmodel} {
 
 		# Call "call_page" with the new enivronment
 		if {[catch {
-			if {$elogfd != ""} {
-				tcl_puts $elogfd "($sock/$addr/[pid]) Debug: [array get myenv] ++ headers = [array get headers]"
-				flush $elogfd
-			}
+if {$elogfd != ""} {
+tcl_puts $elogfd "($sock/$addr/[pid]) Debug: [array get myenv] ++ headers = [array get headers]"
+flush $elogfd
+}
 
 			if {$pmodel == "flat"} {
 				call_page [array get myenv] 1
@@ -1294,10 +1393,16 @@ proc rivet_cgi_server_request_data {sock addr hostport logfd elogfd pmodel} {
 
 		# Cleanup
 		unset sockinfo
-		array set sockinfo {}
+		set sockinfo(state) NEW
 
-		# Tell the event loop that we're done here.
-		set ::rivetstarkit::finished($sock) 1
+		if {$headers(CONNECTION) != "keep-alive"} {
+			# Tell the event loop that we're done here.
+			set ::rivetstarkit::finished($sock) 1
+		} else {
+			fconfigure $sock -buffering line
+			fconfigure $sock -translation crlf
+			fileevent $sock readable [list rivet_cgi_server_request_data $sock $addr $hostport $logfd $elogfd $pmodel]
+		}
 	}
  
 	set ::rivetstarkit::sockinfo($sock) [array get sockinfo]
@@ -1336,7 +1441,7 @@ if {![info exists ::env(GATEWAY_INTERFACE)]} {
 			set rivet_cgi_server_logfile $options(--logfile)
 			set rivet_cgi_server_errorlogfile $options(--errorlog)
 			set rivet_cgi_server_maxthreads $options(--maxthreads)
-			set rivet_cgi_server_sslopts [list certfile $options(--sslcert) keyfile $options(--sslkey) cafile $options(--sslcafile) cadir $options(--sslcadir) require $options(--sslreqcert)]
+			set rivet_cgi_server_sslopts [list certfile $options(--sslcert) keyfile $options(--sslkey) cafile $options(--sslcafile) cadir $options(--sslcadir) request $options(--sslreqcert)]
 
 			rivet_cgi_server $rivet_cgi_server_addr $rivet_cgi_server_port $rivet_cgi_server_fg $rivet_cgi_server_init $rivet_cgi_server_logfile $rivet_cgi_server_errorlogfile $rivet_cgi_server_maxthreads $rivet_cgi_server_sslopts
 
@@ -1352,7 +1457,7 @@ if {![info exists ::env(GATEWAY_INTERFACE)]} {
 			exit 0
 		}
 		"--version" {
-			tcl_puts "RivetStarkit version 0.5.0.2574"
+			tcl_puts "RivetStarkit version 0.5.0.2634"
 			exit 0
 		}
 		default {
